/**************************************************************************
*  This file is part of the TAL project (Tiny Abstraction Layer)
*
*  Copyright (c) 2016-2017 by Michael Fischer (www.emb4fun.de).
*  All rights reserved.
*
*  Redistribution and use in source and binary forms, with or without 
*  modification, are permitted provided that the following conditions 
*  are met:
*  
*  1. Redistributions of source code must retain the above copyright 
*     notice, this list of conditions and the following disclaimer.
*  2. Redistributions in binary form must reproduce the above copyright
*     notice, this list of conditions and the following disclaimer in the 
*     documentation and/or other materials provided with the distribution.
*  3. Neither the name of the author nor the names of its contributors may 
*     be used to endorse or promote products derived from this software 
*     without specific prior written permission.
*
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
*  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
*  OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
*  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
*  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
*  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
*  SUCH DAMAGE.
*
***************************************************************************
*  History:
*
*  27.02.2016  mifi  First Version, tested with an Embedded Artists 
*                    LPC1788 Developer's Kit.
*  18.03.2017  mifi  Complete rework, use structure form LPC2468 port.
*                    Acceptance filter does not work correct from LPCopen.
**************************************************************************/
#define __TALCPU_CAN_C__

/*=======================================================================*/
/*  Include                                                              */
/*=======================================================================*/
#include <string.h>
#include "tal.h"

/*=======================================================================*/
/*  All Structures and Common Constants                                  */
/*=======================================================================*/

typedef struct
{
   volatile uint32_t MOD;
   volatile uint32_t CMR;
   volatile uint32_t GSR;
   volatile uint32_t ICR;
   volatile uint32_t IER;
   volatile uint32_t BTR;
   volatile uint32_t EWL;
   volatile uint32_t SR;

   volatile uint32_t RFS;
   volatile uint32_t RID;
   volatile uint32_t RDA;
   volatile uint32_t RDB;

   volatile uint32_t TFI1;
   volatile uint32_t TID1;
   volatile uint32_t TDA1;
   volatile uint32_t TDB1;

   volatile uint32_t TFI2;
   volatile uint32_t TID2;
   volatile uint32_t TDA2;
   volatile uint32_t TDB2;

   volatile uint32_t TFI3;
   volatile uint32_t TID3;
   volatile uint32_t TDA3;
   volatile uint32_t TDB3;
} CAN_TypeDef;

#define CAN_SJW_1tq     0
#define CAN_SJW_2tq     1
#define CAN_SJW_3tq     2
#define CAN_SJW_4tq     3

#define CAN_TSEG1_1tq   0
#define CAN_TSEG1_2tq   1
#define CAN_TSEG1_3tq   2
#define CAN_TSEG1_4tq   3
#define CAN_TSEG1_5tq   4
#define CAN_TSEG1_6tq   5
#define CAN_TSEG1_7tq   6
#define CAN_TSEG1_8tq   7
#define CAN_TSEG1_9tq   8
#define CAN_TSEG1_10tq  9
#define CAN_TSEG1_11tq  10
#define CAN_TSEG1_12tq  11
#define CAN_TSEG1_13tq  12
#define CAN_TSEG1_14tq  13
#define CAN_TSEG1_15tq  14
#define CAN_TSEG1_16tq  15

#define CAN_TSEG2_1tq   0
#define CAN_TSEG2_2tq   1
#define CAN_TSEG2_3tq   2
#define CAN_TSEG2_4tq   3
#define CAN_TSEG2_5tq   4
#define CAN_TSEG2_6tq   5
#define CAN_TSEG2_7tq   6
#define CAN_TSEG2_8tq   7

/*
 * Some CAN timing calculators:
 * http://www.bittiming.can-wiki.info
 * http://www.port.de/de/bit-timing.html
 */ 

/*
 * BTR defines for 60MHz clock, Sample Point 86.7% (15tq)
 */
#define BTR_1000K_60MHZ ((CAN_SJW_1tq<<14) | (CAN_TSEG2_2tq<<20) | (CAN_TSEG1_12tq<<16) | (  4-1))
#define BTR_500K_60MHZ  ((CAN_SJW_1tq<<14) | (CAN_TSEG2_2tq<<20) | (CAN_TSEG1_12tq<<16) | (  8-1))
#define BTR_250K_60MHZ  ((CAN_SJW_1tq<<14) | (CAN_TSEG2_2tq<<20) | (CAN_TSEG1_12tq<<16) | ( 16-1))
#define BTR_125K_60MHZ  ((CAN_SJW_1tq<<14) | (CAN_TSEG2_2tq<<20) | (CAN_TSEG1_12tq<<16) | ( 32-1))
#define BTR_100K_60MHZ  ((CAN_SJW_1tq<<14) | (CAN_TSEG2_2tq<<20) | (CAN_TSEG1_12tq<<16) | ( 40-1))
#define BTR_50K_60MHZ   ((CAN_SJW_1tq<<14) | (CAN_TSEG2_2tq<<20) | (CAN_TSEG1_12tq<<16) | ( 80-1))
#define BTR_20K_60MHZ   ((CAN_SJW_1tq<<14) | (CAN_TSEG2_2tq<<20) | (CAN_TSEG1_12tq<<16) | (200-1))
#define BTR_10K_60MHZ   ((CAN_SJW_1tq<<14) | (CAN_TSEG2_2tq<<20) | (CAN_TSEG1_12tq<<16) | (400-1))
#define BTR_5K_60MHZ    ((CAN_SJW_1tq<<14) | (CAN_TSEG2_2tq<<20) | (CAN_TSEG1_12tq<<16) | (800-1))


#define CxMOD_RM        C1MOD_RM
#define CxCMR_TR        C1CMR_TR
#define CxCMR_RRB       C1CMR_RRB
#define CxCMR_STB1      C1CMR_STB1
#define CxICR_RI        C1ICR_RI
#define CxICR_TI1       C1ICR_TI1
#define CxICR_EI        C1ICR_EI
#define CxICR_DOI       C1ICR_DOI
#define CxIER_RIE       C1IER_RIE
#define CxIER_TIE1      C1IER_TIE1
#define CxIER_EIE       C1IER_EIE
#define CxIER_DOIE      C1IER_DOIE
#define CxRFS_FF        C1RFS_FF
#define CxRFS_DLC_BIT   C1RFS_DLC_BIT
#define CxTFI1_FF       C1TFI1_FF   

#define STAT_INC(_x)    if (_x < UINT32_MAX) _x++


#if (_TALCPU_CAN_USE_AF >= 1)

/*
 * Filter defines
 */
#define FILTER_ID_MASK        0x1FFFFFFF
#define FILTER_IN_USE_FLAG    0x80000000
#define FILTER_IS_EXT_FLAG    0x40000000

#define FILTER_ID(_x)         (_x & FILTER_ID_MASK)
#define IS_FILTER_IN_USE(_x)  (_x & FILTER_IN_USE_FLAG)
#define IS_FILTER_EXT(_x)     (_x & FILTER_IS_EXT_FLAG)


#if 0 /* Still available in can_17xx_40xx.h */

#define CAN_STD_ENTRY_ID_MASK    (0x7FF)
#define CAN_EXT_ENTRY_ID_MASK    (0x1FFFFFFF)

/*
 * Standard ID Entry structure
 */
typedef struct
{
   uint8_t CtrlNo;               /* Controller Number: 0 for CAN1 and 1 for CAN2 */
   uint8_t Disable;              /* 0(ENABLE)/1(DISABLE): Response On/Off dynamically */
   uint16_t ID_11;               /* Standard ID, should be 11-bit value */
} CAN_STD_ID_ENTRY_T;

/*
 * Standard ID Range structure
 */
typedef struct
{
   CAN_STD_ID_ENTRY_T LowerID;   /* Lower ID Bound, should be in 11-bit value */
   CAN_STD_ID_ENTRY_T UpperID;   /* Upper ID Bound, should be in 11-bit value */
} CAN_STD_ID_RANGE_ENTRY_T;

/*
 * Extended ID Entry structure
 */
typedef struct
{
   uint8_t CtrlNo;               /* Controller Number: 0 for CAN1 and 1 for CAN2 */
   uint32_t ID_29;               /* Extend ID, shoud be 29-bit value */
} CAN_EXT_ID_ENTRY_T;

/*
 * Extended ID Range structure
 */
typedef struct 
{
   CAN_EXT_ID_ENTRY_T LowerID;   /* Lower ID Bound, should be in 29-bit value */
   CAN_EXT_ID_ENTRY_T UpperID;   /* Upper ID Bound, should be in 29-bit value */
} CAN_EXT_ID_RANGE_ENTRY_T;

/*
 * Acceptance Filter Section Table structure
 */
typedef struct
{
   CAN_STD_ID_ENTRY_T *FullCANSec;        /* The pointer to fullCAN section */
   uint16_t FullCANEntryNum;              /* FullCAN Entry Number */

   CAN_STD_ID_ENTRY_T *SffSec;            /* The pointer to individual Standard ID Section */
   uint16_t SffEntryNum;                  /* Standard ID Entry Number */
   
   CAN_STD_ID_RANGE_ENTRY_T *SffGrpSec;   /* The pointer to  Group Standard ID  Section */
   uint16_t SffGrpEntryNum;               /* Group Standard ID Entry Number */
   
   CAN_EXT_ID_ENTRY_T *EffSec;            /* The pointer to Extended ID Section */
   uint16_t EffEntryNum;                  /* Extended ID Entry Number */
   
   CAN_EXT_ID_RANGE_ENTRY_T *EffGrpSec;   /* The pointer to Group Extended ID Section */
   uint16_t EffGrpEntryNum;               /* Group Extended ID Entry Number */
} CANAF_LUT_T;
#endif

#else
#define UpdateAcceptanceFilter()
#endif
            
/*=======================================================================*/
/*  Definition of all local Data                                         */
/*=======================================================================*/

#if (_TALCPU_CAN_USE_AF >= 1)
/*
 * Acceptance Filter
 */
static CAN_STD_ID_ENTRY_T       AFStdSection[(uint32_t)TAL_CAN_PORT_MAX * CAN_MAX_FILTER_COUNT];
static CAN_EXT_ID_ENTRY_T       AFExtSection[(uint32_t)TAL_CAN_PORT_MAX * CAN_MAX_FILTER_COUNT];
static CAN_STD_ID_RANGE_ENTRY_T AFStdGroupSection[TAL_CAN_PORT_MAX];
static CAN_EXT_ID_RANGE_ENTRY_T AFExtGroupSection[TAL_CAN_PORT_MAX];
static CANAF_LUT_T              AFSections;
#endif

static TAL_CAN_DCB *DCBArray[TAL_CAN_PORT_MAX];

/*=======================================================================*/
/*  Definition of all local Procedures                                   */
/*=======================================================================*/

#if (_TALCPU_CAN_USE_AF >= 1)
/*************************************************************************/
/*  SetAFLUT                                                             */
/*                                                                       */
/*  Set AF Lookup Table.                                                 */
/*                                                                       */
/*  In    : pAFSections                                                  */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void SetAFLUT (CANAF_LUT_T *pAFSections)
{
   uint16_t wIndex;
   uint32_t *pFilter;
   uint16_t  wStdFilter1;
   uint16_t  wStdFilter2;
   uint32_t  dValue;
   uint32_t  dCtrlNo;
   
   /* Disable Acceptance Filter */
   AFMR = AFMR_AccOff;
   
   /*
    * Check Standard filter section, count must be even.
    */
   if (pAFSections->SffEntryNum & 0x01)
   {
      pAFSections->SffEntryNum++;
      wIndex = pAFSections->SffEntryNum - 1;
      
      /* Set additional index */
      pAFSections->SffSec[wIndex].CtrlNo  = 0;
      pAFSections->SffSec[wIndex].ID_11   = 0;
      pAFSections->SffSec[wIndex].Disable = 1;
   }

   pFilter = (uint32_t*)CAN_BASE;
   SFF_sa     = 0; /* No FullCAN section */
   SFF_GRP_sa = 0;
   EFF_sa     = 0;
   EFF_GRP_sa = 0;
   ENDofTable = 0;

   memset((void*)CAN_BASE, 0x00, 0x800);
   
   
   /*
    * Check if Standard section is needed 
    */
   if      (pAFSections->SffEntryNum)
   {
      for(wIndex=0; wIndex<pAFSections->SffEntryNum; wIndex+=2)
      {
         wStdFilter1  = pAFSections->SffSec[wIndex].ID_11;
         wStdFilter1 |= (pAFSections->SffSec[wIndex].Disable ? (1<<12) : 0);
         wStdFilter1 |= (uint16_t)(pAFSections->SffSec[wIndex].CtrlNo << 13);
         
         wStdFilter2  = pAFSections->SffSec[wIndex+1].ID_11;
         wStdFilter2 |= (pAFSections->SffSec[wIndex+1].Disable ? (1<<12) : 0);
         wStdFilter2 |= (uint16_t)(pAFSections->SffSec[wIndex+1].CtrlNo << 13);
         
         *pFilter++ = ((uint32_t)wStdFilter1 << 16) | wStdFilter2;
      }

      SFF_GRP_sa = ((uint32_t)pFilter & 0x3FF);
      EFF_sa     = ((uint32_t)pFilter & 0x3FF);
      EFF_GRP_sa = ((uint32_t)pFilter & 0x3FF);
      ENDofTable = ((uint32_t)pFilter & 0x3FF);
   }
   /* Check if Standard group section is needed */
   else if (pAFSections->SffGrpEntryNum)
   {
      for(wIndex=0; wIndex<pAFSections->SffGrpEntryNum; wIndex++)
      {
         dCtrlNo = pAFSections->SffGrpSec[wIndex].LowerID.CtrlNo;
         
         dValue  = CAN_STD_ENTRY_ID_MASK;
         dValue |= (dCtrlNo<<29);
         dValue |= (dCtrlNo<<14);
         
         *pFilter++ = dValue;
      }
      
      SFF_GRP_sa = SFF_sa;
      EFF_sa     = ((uint32_t)pFilter & 0x3FF);
      EFF_GRP_sa = ((uint32_t)pFilter & 0x3FF);
      ENDofTable = ((uint32_t)pFilter & 0x3FF);
   }


   /*
    * Check if Extended section is needed
    */
   if      (pAFSections->EffEntryNum)
   {
      for(wIndex=0; wIndex<pAFSections->EffEntryNum; wIndex++)
      {
         dCtrlNo = pAFSections->EffSec[wIndex].CtrlNo;
      
         dValue  = pAFSections->EffSec[wIndex].ID_29;
         dValue |= dCtrlNo << 29;
         
         *pFilter++ = dValue;
      }

      EFF_GRP_sa = ((uint32_t)pFilter & 0x3FF);
      ENDofTable = ((uint32_t)pFilter & 0x3FF);
   }
   /* Check if Standard group section is needed */
   else if (pAFSections->EffGrpEntryNum)
   {
      for(wIndex=0; wIndex<pAFSections->EffGrpEntryNum; wIndex++)
      {
         /* Lower */
         dCtrlNo = pAFSections->EffGrpSec[wIndex].LowerID.CtrlNo;
      
         dValue  = pAFSections->EffGrpSec[wIndex].LowerID.ID_29;
         dValue |= dCtrlNo << 29;
         
         *pFilter++ = dValue;
      
         /* Upper */
         dValue  = pAFSections->EffGrpSec[wIndex].UpperID.ID_29;
         dValue |= dCtrlNo << 29;
         
         *pFilter++ = dValue;
      }
      
      EFF_GRP_sa = EFF_sa;
      ENDofTable = ((uint32_t)pFilter & 0x3FF);
   }

   /* Enable Acceptance Filter */
   AFMR &= ~AFMR_AccOff;
   
} /* SetAFLUT */

/*************************************************************************/
/*  SortFilterList                                                       */
/*                                                                       */
/*  Sort the filter list in ascending numerical order.                   */
/*                                                                       */
/*  In    : Filter, FilterSize                                           */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void SortFilterList (uint32_t *Filter, uint16_t FilterSize)
{
   uint32_t TmpFilter;

   uint16_t i;
   uint16_t j;
   
   for (i = 0; i < FilterSize - 1; i++)
   {
      for (j = 0; j < (FilterSize - i) - 1; j++)
      {
         if (FILTER_ID(Filter[j]) > FILTER_ID(Filter[j + 1]))
         {
            TmpFilter     = Filter[j];
            Filter[j]     = Filter[j + 1];
            Filter[j + 1] = TmpFilter;
         }
      }
   }
} /* SortFilterList */

/*************************************************************************/
/*  UpdateAcceptanceFilter                                               */
/*                                                                       */
/*  Update the acceptance filter.                                        */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: TAL_OK                                                       */
/*************************************************************************/
static TAL_RESULT UpdateAcceptanceFilter (void)
{
   TAL_CAN_DCB *pDCB;
   uint8_t      bPort;
   uint16_t     wIndex;
   uint16_t     wStdCount      = 0;
   uint16_t     wExtCount      = 0;
   uint16_t     wStdGroupCount = 0;
   uint16_t     wExtGroupCount = 0;

   for (bPort = 0; bPort < TAL_CAN_PORT_MAX; bPort++)
   {
      pDCB = DCBArray[bPort];
      if (pDCB != NULL)
      {
         SortFilterList(pDCB->HW.FilterArray, CAN_MAX_FILTER_COUNT); 
   
         for (wIndex = 0; wIndex < CAN_MAX_FILTER_COUNT; wIndex++)
         {
            if (IS_FILTER_IN_USE(pDCB->HW.FilterArray[wIndex]))
            {
               if (IS_FILTER_EXT(pDCB->HW.FilterArray[wIndex]))
               {
                  AFExtSection[wExtCount].CtrlNo  = bPort;
                  AFExtSection[wExtCount].ID_29   = FILTER_ID(pDCB->HW.FilterArray[wIndex]);
                  
                  wExtCount++;
               }
               else
               {
                  AFStdSection[wStdCount].CtrlNo  = bPort;
                  AFStdSection[wStdCount].Disable = 0;
                  AFStdSection[wStdCount].ID_11   = (uint16_t)FILTER_ID(pDCB->HW.FilterArray[wIndex]);
                  
                  wStdCount++;
               }   
            }
         }
         
         if (TAL_TRUE == pDCB->HW.bFilterAllStd)
         {
            AFStdGroupSection[wStdGroupCount].LowerID.CtrlNo  = bPort;
            AFStdGroupSection[wStdGroupCount].LowerID.Disable = 0;
            AFStdGroupSection[wStdGroupCount].LowerID.ID_11   = 0;
            
            AFStdGroupSection[wStdGroupCount].UpperID.CtrlNo  = bPort;
            AFStdGroupSection[wStdGroupCount].UpperID.Disable = 0;
            AFStdGroupSection[wStdGroupCount].UpperID.ID_11   = CAN_STD_ENTRY_ID_MASK;
            
            wStdGroupCount++;
            
            /*
             * Explicit Standard Frame Format Identifier Section not needed.
             */
            wStdCount = 0;
         } 

         if (TAL_TRUE == pDCB->HW.bFilterAllExt)
         {
            AFExtGroupSection[wExtGroupCount].LowerID.CtrlNo  = bPort;
            AFExtGroupSection[wExtGroupCount].LowerID.ID_29   = 0;
            
            AFExtGroupSection[wExtGroupCount].UpperID.CtrlNo  = bPort;
            AFExtGroupSection[wExtGroupCount].UpperID.ID_29   = CAN_EXT_ENTRY_ID_MASK;
            
            wExtGroupCount++;            
            
            /*
             * Explicit Extended Frame Format Identifier Section not needed.
             */
            wExtCount = 0;            
         } 
         
      } /* end if (pDCB != NULL) */ 
   } /* end for (bPort = 0; bPort < TAL_CAN_PORT_MAX; bPort++) */
   

   /*
    * Set Acceptance Filter sections
    */
   AFSections.FullCANSec      = NULL;
   AFSections.FullCANEntryNum = 0;
   
   AFSections.SffSec          = AFStdSection;
   AFSections.SffEntryNum     = wStdCount;
   
   AFSections.SffGrpSec       = AFStdGroupSection;
   AFSections.SffGrpEntryNum  = wStdGroupCount;
   
   AFSections.EffSec          = AFExtSection;
   AFSections.EffEntryNum     = wExtCount;
   
   AFSections.EffGrpSec       = AFExtGroupSection;
   AFSections.EffGrpEntryNum  = wExtGroupCount;
   
   SetAFLUT(&AFSections);   

   return(TAL_OK);
} /* UpdateAcceptanceFilter */

/*************************************************************************/
/*  CANIdentRegister                                                     */
/*                                                                       */
/*  Register a CAN identifier in the acceptance filter.                  */
/*                                                                       */
/*  In    : pDCB, dIdentifier, Type                                      */
/*  Out   : none                                                         */
/*  Return: TAL_OK / error cause                                         */
/*************************************************************************/
static TAL_RESULT CANIdentRegister (TAL_CAN_DCB   *pDCB,
                                    uint32_t       dIdentifier,
                                    TAL_CAN_TYPE_ID Type)
{
   TAL_RESULT  Error = TAL_ERROR;
   uint16_t   wIndex;
   uint32_t  *FilterArray = pDCB->HW.FilterArray;
   uint8_t    bRegsiterOK = TAL_FALSE;
   
   
   /* 
    * Handle special case TAL_CAN_ID_RX_ALL, all identifier
    * from type STD or EXT should be received.
    */
   if (TAL_CAN_ID_RX_ALL == dIdentifier)
   {
      /* Test if STD or EXT ID */
      if (TAL_CAN_TYPE_STD_ID == Type)
      {
         pDCB->HW.bFilterAllStd = TAL_TRUE;
         
         /* Clear all other STD identifier */
         for (wIndex = 0; wIndex < CAN_MAX_FILTER_COUNT; wIndex++)
         {
            if ((FilterArray[wIndex] & FILTER_IN_USE_FLAG)        &&
                (0 == (FilterArray[wIndex] & FILTER_IS_EXT_FLAG)) )
            {
               FilterArray[wIndex] = 0;
            }
         }
      }
      else
      {
         pDCB->HW.bFilterAllExt = TAL_TRUE;

         /* Clear all other EXT identifier */
         for (wIndex = 0; wIndex < CAN_MAX_FILTER_COUNT; wIndex++)
         {
            if ((FilterArray[wIndex] & FILTER_IN_USE_FLAG) &&
                (FilterArray[wIndex] & FILTER_IS_EXT_FLAG))
            {
               FilterArray[wIndex] = 0;
            }
         }
      } 

      Error = UpdateAcceptanceFilter();    
      goto IdentRegisterEnd;     /*lint !e801*/
   } /* end if (TAL_CAN_ID_RX_ALL == dIdentifier) */


   /*
    * Check for STD or EXT identifier
    */   
   if (TAL_CAN_TYPE_STD_ID == Type)
   {
      /* Check valid range */
      if (dIdentifier > CAN_STD_ENTRY_ID_MASK)
      {
         Error = TAL_ERR_CAN_ID_RANGE;
         goto IdentRegisterEnd;  /*lint !e801*/
      } 

      if (TAL_TRUE == pDCB->HW.bFilterAllStd)
      {
         Error = TAL_ERR_CAN_ID_USED;
         goto IdentRegisterEnd;  /*lint !e801*/
      }    
   }
   else
   {
      /* Check valid range */
      if (dIdentifier > CAN_EXT_ENTRY_ID_MASK)
      {
         Error = TAL_ERR_CAN_ID_RANGE;
         goto IdentRegisterEnd;  /*lint !e801*/
      }
      
      if (TAL_TRUE == pDCB->HW.bFilterAllExt)
      {
         Error = TAL_ERR_CAN_ID_USED;
         goto IdentRegisterEnd;  /*lint !e801*/
      }    

      dIdentifier |= FILTER_IS_EXT_FLAG; 
   }
   
   dIdentifier |= FILTER_IN_USE_FLAG;
   
   /*
    * Check if this identifier is registered
    */
   for (wIndex = 0; wIndex < CAN_MAX_FILTER_COUNT; wIndex++)
   {
      if (FilterArray[wIndex] == dIdentifier)
      {
         Error = TAL_ERR_CAN_ID_USED;
         goto IdentRegisterEnd;  /*lint !e801*/
      }
   }
   
   
   /*
    * At this point the identifier is not registered.
    * Find a free entry which can be used.
    */
   for (wIndex = 0; wIndex < CAN_MAX_FILTER_COUNT; wIndex++)
   {
      if (0 == IS_FILTER_IN_USE(FilterArray[wIndex]))
      {
         FilterArray[wIndex] = dIdentifier;
         bRegsiterOK         = TAL_TRUE;
         break;
      }
   }      
   if (TAL_FALSE == bRegsiterOK)
   {
      Error = TAL_ERR_CAN_ID_NO_ENTRY;
      goto IdentRegisterEnd;  /*lint !e801*/
   }


   /*
    * At this point an update of the Acceptance Filter is needed
    */
   Error = UpdateAcceptanceFilter();    
   
IdentRegisterEnd:   
   return(Error);
} /* CANIdentRegister */   

/*************************************************************************/
/*  CANIdentDeRegister                                                   */
/*                                                                       */
/*  DeRegister a CAN identifier from the acceptance filter.              */
/*                                                                       */
/*  In    : pDCB, dIdentifier, Type                                      */
/*  Out   : none                                                         */
/*  Return: TAL_OK / error cause                                         */
/*************************************************************************/
static TAL_RESULT CANIdentDeRegister (TAL_CAN_DCB   *pDCB,
                                      uint32_t       dIdentifier,
                                      TAL_CAN_TYPE_ID Type)
{
   TAL_RESULT  Error = TAL_ERROR;     
   uint16_t   wIndex;
   uint32_t  *FilterArray   = pDCB->HW.FilterArray;
   uint8_t    bDeRegsiterOK = TAL_FALSE;
  

   /* 
    * Handle special case TAL_CAN_ID_RX_ALL, all identifier
    * from type STD or EXT should be received.
    */
   if (TAL_CAN_ID_RX_ALL == dIdentifier)
   {
      /* Test if STD or EXT ID */
      if (TAL_CAN_TYPE_STD_ID == Type)
      {
         pDCB->HW.bFilterAllStd = TAL_FALSE;
      }
      else
      {
         pDCB->HW.bFilterAllExt = TAL_FALSE;
      }   

      Error = UpdateAcceptanceFilter();    
      goto IdentDeRegisterEnd;      /*lint !e801*/
   } /* end if (TAL_CAN_ID_RX_ALL == dIdentifier) */

  
   /*
    * Check for STD or EXT identifier
    */   
   if (TAL_CAN_TYPE_STD_ID == Type)
   {
      /* Check valid range */
      if (dIdentifier > CAN_STD_ENTRY_ID_MASK)
      {
         Error = TAL_ERR_CAN_ID_RANGE;
         goto IdentDeRegisterEnd;   /*lint !e801*/
      }
   }
   else
   {
      /* Check valid range */
      if (dIdentifier > CAN_EXT_ENTRY_ID_MASK)
      {
         Error = TAL_ERR_CAN_ID_RANGE;
         goto IdentDeRegisterEnd;   /*lint !e801*/
      }
      
      dIdentifier |= FILTER_IS_EXT_FLAG; 
   }
   
   dIdentifier |= FILTER_IN_USE_FLAG;
   
   /*
    * Check if this identifier is registered,
    * and remove it from the list if available.
    */
   for (wIndex = 0; wIndex < CAN_MAX_FILTER_COUNT; wIndex++)
   {
      if (FilterArray[wIndex] == dIdentifier)
      {
         FilterArray[wIndex] = 0;
         bDeRegsiterOK       = TAL_TRUE;
         break;
      }
   }
   if (TAL_FALSE == bDeRegsiterOK)
   {
      Error = TAL_ERR_CAN_ID_NOT_USED;
      goto IdentDeRegisterEnd;   /*lint !e801*/
   }

   /*
    * At this point an update of the Acceptance Filter is needed
    */
   Error = UpdateAcceptanceFilter();    

IdentDeRegisterEnd:   
   return(Error);
} /* CANIdentDeRegister */   
#else
/*************************************************************************/
/*  CANIdentRegister                                                     */
/*                                                                       */
/*  Register a CAN identifier in the bit array.                          */
/*                                                                       */
/*  In    : pDCB, dIdentifier, Type                                      */
/*  Out   : none                                                         */
/*  Return: TAL_OK / error cause                                         */
/*************************************************************************/
static TAL_RESULT CANIdentRegister (TAL_CAN_DCB   *pDCB,
                                    uint32_t       dIdentifier,
                                    TAL_CAN_TYPE_ID Type)
{
   TAL_RESULT  Error = TAL_ERROR;
   uint8_t    bIndex;
   uint8_t    bBitPos; 
   
   /*
    * Only standard identifier will be supported.
    */
   if (TAL_CAN_TYPE_STD_ID == Type)
   {
      if      (dIdentifier <= CAN_STD_ENTRY_ID_MASK)
      {
         Error = TAL_OK;
      
         bIndex  = (uint8_t)(dIdentifier / 32);
         bBitPos = (uint8_t)(dIdentifier % 32);
      
         TAL_CPU_DISABLE_ALL_INTS();
      
         pDCB->HW.FilterArray[bIndex] |= (1u << bBitPos);
      
         TAL_CPU_ENABLE_ALL_INTS();
      }
      else if (TAL_CAN_ID_RX_ALL == dIdentifier)
      {
         Error = TAL_OK;
      
         TAL_CPU_DISABLE_ALL_INTS();
      
         memset(pDCB->HW.FilterArray, 0xFF, sizeof(pDCB->HW.FilterArray));
      
         TAL_CPU_ENABLE_ALL_INTS();
      }
   }

   return(Error);
} /* CANIdentRegister */

/*************************************************************************/
/*  CANIdentDeRegister                                                   */
/*                                                                       */
/*  DeRegister a CAN identifier from the bit array.                      */
/*                                                                       */
/*  In    : pDCB, dIdentifier, Type                                      */
/*  Out   : none                                                         */
/*  Return: TAL_OK / error cause                                         */
/*************************************************************************/
static TAL_RESULT CANIdentDeRegister (TAL_CAN_DCB   *pDCB,
                                      uint32_t       dIdentifier,
                                      TAL_CAN_TYPE_ID Type)
{
   TAL_RESULT  Error = TAL_ERROR;
   uint8_t    bIndex;
   uint8_t    bBitPos; 
   
   /*
    * Only standard identifier will be supported.
    */
   if (TAL_CAN_TYPE_STD_ID == Type)
   {
      if      (dIdentifier <= CAN_STD_ENTRY_ID_MASK)
      {
         Error = TAL_OK;
      
         bIndex  = (uint8_t)(dIdentifier / 32);
         bBitPos = (uint8_t)(dIdentifier % 32);
      
         TAL_CPU_DISABLE_ALL_INTS();
      
         pDCB->HW.FilterArray[bIndex] &= ~(1u << bBitPos);
      
         TAL_CPU_ENABLE_ALL_INTS();
      }
      else if (TAL_CAN_ID_RX_ALL == dIdentifier)
      {
         Error = TAL_OK;
      
         TAL_CPU_DISABLE_ALL_INTS();
      
         memset(pDCB->HW.FilterArray, 0x00, sizeof(pDCB->HW.FilterArray));
      
         TAL_CPU_ENABLE_ALL_INTS();
      }
   }

   return(Error);
} /* CANIdentDeRegister */
#endif /* (_TALCPU_CAN_USE_AF >= 1) */

/*************************************************************************/
/*  SendObject                                                           */
/*                                                                       */
/*  Try to send the given Object.                                        */
/*                                                                       */
/*  In    : pDCB, pCANx, pCANObject                                      */
/*  Out   : none                                                         */
/*  Return: TAL_OK                                                       */
/*************************************************************************/
static TAL_RESULT SendObject (TAL_CAN_DCB *pDCB, CAN_TypeDef *pCANx, TAL_CAN_OBJECT *pCANObject)
{
   uint32_t dTFI;
   uint32_t dTID;
   uint32_t dTDA;
   uint32_t dTDB;
   
   /* Setup Identifier */
   dTFI = ((uint32_t)pCANObject->bDLC << 16);
   dTID = pCANObject->dIdentifier;
   if (pCANObject->bFlags & TAL_CAN_OBJECT_FLAGS_EXT_ID)
   {
      dTFI |= CxTFI1_FF; 
   }   
   
   /* Setup data */ 
   dTDA  = pCANObject->Data[3];
   dTDA  = dTDA << 8;
   dTDA |= pCANObject->Data[2];
   dTDA  = dTDA << 8;
   dTDA |= pCANObject->Data[1];
   dTDA  = dTDA << 8;
   dTDA |= pCANObject->Data[0];
   
   dTDB  = pCANObject->Data[7];
   dTDB  = dTDB << 8;
   dTDB |= pCANObject->Data[6];
   dTDB  = dTDB << 8;
   dTDB |= pCANObject->Data[5];
   dTDB  = dTDB << 8;
   dTDB |= pCANObject->Data[4];

   /* Setup CAN1-TX register */   
   pCANx->TFI1 = dTFI;
   pCANx->TID1 = dTID;
   pCANx->TDA1 = dTDA;
   pCANx->TDB1 = dTDB;

   /* Send message */
   pCANx->CMR = (CxCMR_STB1 | CxCMR_TR);
   
   STAT_INC(pDCB->HW.Stat.dTXCount);

   return(TAL_OK);   
} /* SendObject */

/*************************************************************************/
/*  TX_IRQHandler                                                        */
/*                                                                       */
/*  This is the generic TX IRQHandler.                                   */
/*                                                                       */
/*  In    : pDCB                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void TX_IRQHandler (TAL_CAN_DCB *pDCB)
{
   TAL_RESULT     Error;
   TAL_CAN_OBJECT CANObject;
   CAN_TypeDef  *pCANx;

   /* Get CAN pointer */
   pCANx = (CAN_TypeDef *)pDCB->HW.dBaseAddress;

   /* 
    * Check for TX interrupt, but only if it is anabled 
    */
   if (pCANx->IER & CxIER_TIE1)
   {
      /* Read Object from the ring buffer */
      Error = tal_MISCRingGet(&pDCB->TxRing, (uint8_t*)&CANObject);
      if (Error != TAL_OK)
      {
         /* Ups, no data available, disable interrupt */
         pCANx->IER &= ~CxIER_TIE1;
      }
      else
      {
         /* Send Object */
         SendObject(pDCB, pCANx, &CANObject); 
      }   
   } /* end "TX interrupt */
   
   (void)pDCB;
} /* TX_IRQHandler */

/*************************************************************************/
/*  RX_IRQHandler                                                        */
/*                                                                       */
/*  This is the generic RX IRQHandler.                                   */
/*                                                                       */
/*  In    : pDCB                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void RX_IRQHandler (TAL_CAN_DCB *pDCB)
{
   TAL_RESULT     Error;
   TAL_CAN_OBJECT CANObject;
   CAN_TypeDef  *pCANx;
   uint32_t      dRFS;
   uint32_t      dRID;
   uint32_t      dRDA;
   uint32_t      dRDB;
   uint8_t       bIndex = 0;
   uint8_t       bBitPos = 0;
   
   /* Set timestamp, if supported */
   if (NULL == pDCB->GetRxTimestamp)
   { 
      CANObject.dTimestamp = 0;
   }
   else
   {
      CANObject.dTimestamp = pDCB->GetRxTimestamp();
   }   

   /* Get CAN pointer */
   pCANx = (CAN_TypeDef*)pDCB->HW.dBaseAddress;

   /* Get CAN message */
   dRFS = pCANx->RFS;
   dRID = pCANx->RID;
   dRDA = pCANx->RDA;
   dRDB = pCANx->RDB;

   /* Release Receive Buffer */
   pCANx->CMR = CxCMR_RRB;

   /* Get Identifier */
   if (dRFS & CxRFS_FF)
   {
     /* Only supported in Acceptance Filter mode */
#if (_TALCPU_CAN_USE_AF >= 1)
      /* 29 bit Identifier */
      CANObject.bFlags      = TAL_CAN_OBJECT_FLAGS_EXT_ID;
      CANObject.dIdentifier = (dRID & 0x1FFFFFFF);
#else
      return;
#endif         
   }
   else
   {
      /* 11 bit Identifier */
      CANObject.bFlags      = TAL_CAN_OBJECT_FLAGS_STD_ID;
      CANObject.dIdentifier = (dRID & 0x7FF);
   }   
   
#if (_TALCPU_CAN_USE_AF == 0)
   bIndex  = (uint8_t)(CANObject.dIdentifier / 32);
   bBitPos = (uint8_t)(CANObject.dIdentifier % 32);
   
   /* Check if we should receive this identifier */
   if (0 == (pDCB->HW.FilterArray[bIndex] & (1u << bBitPos)))
   {
      return;
   } 
#endif

   /* Get DLC */
   CANObject.bDLC = (dRFS >> CxRFS_DLC_BIT) & 0x0F;
   if (CANObject.bDLC > 7)
   {
      CANObject.bDLC = 8;
   }
   
   /* Get data */
   CANObject.Data[0] = (uint8_t)(dRDA & 0xFF);
   dRDA = dRDA >> 8;
   CANObject.Data[1] = (uint8_t)(dRDA & 0xFF);
   dRDA = dRDA >> 8;
   CANObject.Data[2] = (uint8_t)(dRDA & 0xFF);
   dRDA = dRDA >> 8;
   CANObject.Data[3] = (uint8_t)(dRDA & 0xFF);

   CANObject.Data[4] = (uint8_t)(dRDB & 0xFF);
   dRDB = dRDB >> 8;
   CANObject.Data[5] = (uint8_t)(dRDB & 0xFF);
   dRDB = dRDB >> 8;
   CANObject.Data[6] = (uint8_t)(dRDB & 0xFF);
   dRDB = dRDB >> 8;
   CANObject.Data[7] = (uint8_t)(dRDB & 0xFF);
   
   /* If we have no overflow... */
   if (TAL_FALSE == pDCB->bRxOverflow)
   {
      /* ... put it into the ring buffer */
      Error = tal_MISCRingAdd(&pDCB->RxRing, (uint8_t*)&CANObject);
      if (TAL_OK == Error)
      {
         /* Signal counting semaphore */
         OSSemaSignalFromInt(&pDCB->RxRdySema);
      }
      else
      {
         /* Ups, overflow */
         pDCB->bRxOverflow = TAL_OK;
      }
   } 
   
   STAT_INC(pDCB->HW.Stat.dRXCount);
   
   (void)bIndex;
   (void)bBitPos;

} /* RX_IRQHandler */

/*=======================================================================*/
/*  All code exported                                                    */
/*=======================================================================*/

/*************************************************************************/
/*  CAN_IRQHandler                                                       */
/*                                                                       */
/*  This is the Cortex CAN IRQ handler.                                  */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void CAN_IRQHandler (void)
{
   uint32_t dStatus1 = C1ICR;
   uint32_t dStatus2 = C2ICR;

   TAL_CPU_IRQ_ENTER();

   /*
    * RX Interrupt
    */
   if (dStatus1 & CxICR_RI)
   {
      RX_IRQHandler(DCBArray[TAL_CAN_PORT_1]);
   }
   
   if (dStatus2 & CxICR_RI)
   {
      RX_IRQHandler(DCBArray[TAL_CAN_PORT_2]);
   }

   /*
    * TX Interrupt
    */
   if ((dStatus1 & CxICR_TI1) && (CAN1IER & CxIER_TIE1))
   {
      TX_IRQHandler(DCBArray[TAL_CAN_PORT_1]);
   }
   
   if ((dStatus2 & CxICR_TI1) && (CAN2IER & CxIER_TIE1))
   {
      TX_IRQHandler(DCBArray[TAL_CAN_PORT_2]);
   }
   
   /*
    * Error Interrupt
    */
   if (dStatus1 & CxICR_EI)
   {
      /* Reset and restart CAN */
      C1MOD  = C1MOD_RM;
      C1GSR  = 0;
      C1MOD  = 0;
      C1IER &= ~C1IER_TIE1;
   }
   
   if (dStatus2 & CxICR_EI)
   {
      /* Reset and restart CAN */
      C2MOD  = C2MOD_RM;
      C2GSR  = 0;
      C2MOD  = 0;
      C2IER &= ~C2IER_TIE1;
   }
   
   /*
    * Check DOS status for CAN1 and CAN2
    */
   if (dStatus1 & CxICR_DOI)
   {
      STAT_INC(DCBArray[TAL_CAN_PORT_1]->HW.Stat.dOverunCount);
   }
   if (dStatus2 & CxICR_DOI)
   {
      STAT_INC(DCBArray[TAL_CAN_PORT_2]->HW.Stat.dOverunCount);
   }
   
   TAL_CPU_IRQ_EXIT();
} /* CAN_IRQHandler */

/*************************************************************************/
/*  cpu_CANInit                                                          */
/*                                                                       */
/*  Prepare the hardware for use by the Open function later. Set the HW  */
/*  information depending of ePort and "enable" the CAN port.            */
/*                                                                       */
/*  In    : pDCB                                                         */
/*  Out   : none                                                         */
/*  Return: TAL_OK / error cause                                         */
/*************************************************************************/
TAL_RESULT cpu_CANInit (TAL_CAN_DCB *pDCB)
{
   TAL_RESULT      Error     = TAL_ERR_CAN_PORT_RANGE;
   TAL_CAN_HW    *pHW        = &pDCB->HW;
      
   switch (pDCB->ePort)
   {
      case TAL_CAN_PORT_1:
      {
         Error = tal_BoardEnableCAN1();
         if (TAL_OK == Error)
         {
            DCBArray[TAL_CAN_PORT_1] = pDCB;

            /* Clear HW info first */
            memset(pHW, 0x00, sizeof(TAL_CAN_HW));

            pHW->dBaseAddress = CAN1_BASE;
            
            C1MOD = CxMOD_RM; /* CAN operation is disabled */
            C1IER = 0;        /* Disable all interrupts */

            /* Set priority and function */
            tal_CPUIrqSetPriority(CAN_IRQn, CAN_PRIO);
            tal_CPUIrqEnable(CAN_IRQn);
         }
         break;
      } /* TAL_CAN_PORT_1 */
   
      case TAL_CAN_PORT_2:
      {
         Error = tal_BoardEnableCAN2();
         if (TAL_OK == Error)
         {
            DCBArray[TAL_CAN_PORT_2] = pDCB;

            /* Clear HW info first */
            memset(pHW, 0x00, sizeof(TAL_CAN_HW));

            pHW->dBaseAddress = CAN2_BASE;
            
            C2MOD = C2MOD_RM; /* CAN operation is disabled */
            C2IER = 0;        /* Disable all interrupts */

            /* Set priority and function */
            tal_CPUIrqSetPriority(CAN_IRQn, CAN_PRIO);
            tal_CPUIrqEnable(CAN_IRQn);
         }
         break;
      } /* TAL_CAN_PORT_2 */
      
      default:
      {
         /* Do nothing */
         Error = TAL_ERR_CAN_PORT_RANGE;
         break;
      }
   } /* end switch (pDCB->ePort) */
   
   return(Error);
} /* cpu_CANInit */

/*************************************************************************/
/*  cpu_CANOpen                                                          */
/*                                                                       */
/*  Open the CAN port.                                                   */
/*                                                                       */
/*  In    : pDCB                                                         */
/*  Out   : none                                                         */
/*  Return: TAL_OK / error cause                                         */
/*************************************************************************/
TAL_RESULT cpu_CANOpen (TAL_CAN_DCB *pDCB)
{
   TAL_RESULT  Error = TAL_ERROR;
   CAN_TypeDef   *pCANx;
   uint32_t       dBTR = 0;
   static uint8_t bFirstCall = 1;
   
   /* Get CAN pointer */
   pCANx = (CAN_TypeDef*)pDCB->HW.dBaseAddress;

   /* 
    * Check baud rate 
    */
   if (60000000L == tal_CPUGetFrequencyAPB())
   {
      switch (pDCB->Settings.dBitRate)
      {
         case TAL_CAN_BR_1000K: dBTR = BTR_1000K_60MHZ; break;
         case TAL_CAN_BR_500K:  dBTR = BTR_500K_60MHZ;  break;
         case TAL_CAN_BR_250K:  dBTR = BTR_250K_60MHZ;  break;
         case TAL_CAN_BR_125K:  dBTR = BTR_125K_60MHZ;  break;
         case TAL_CAN_BR_100K:  dBTR = BTR_100K_60MHZ;  break;
         case TAL_CAN_BR_50K:   dBTR = BTR_50K_60MHZ;   break;
         case TAL_CAN_BR_20K:   dBTR = BTR_20K_60MHZ;   break;
         case TAL_CAN_BR_10K:   dBTR = BTR_10K_60MHZ;   break;
         case TAL_CAN_BR_5K:    dBTR = BTR_5K_60MHZ;    break;
      
         default:
         {
            Error = TAL_ERR_CAN_BAUDRATE;
            goto CANOpenEnd;  /*lint !e801*/
            break;   /*lint !e527*/
         }
      } /* end switch (pDCB->Settings.dBaudrate) */
   }
   else
   {
      Error = TAL_ERR_CAN_BAUDRATE;
      goto CANOpenEnd;  /*lint !e801*/
   }

   /*
    * "Clear" the Acceptance Filter for the first call
    */   
   if (1 == bFirstCall)
   {
      bFirstCall = 0;
  
      /* Disable Acceptance Filter */
      AFMR = AFMR_AccOff;

      SFF_sa     = 0;
      SFF_GRP_sa = 0;
      EFF_sa     = 0;
      EFF_GRP_sa = 0;
      ENDofTable = 0;
   
      memset((void*)CAN_BASE, 0x00, 0x800);

      /* Enable Acceptance Filter */
      AFMR &= ~AFMR_AccOff;
   }

   pCANx->MOD = CxMOD_RM;     /* CAN operation is disabled */
   pCANx->IER = 0;            /* Disable all interrupts */
   pCANx->GSR = 0;            /* Reset error counter when CANxMOD is in reset */   
   pCANx->BTR = dBTR;
   pCANx->MOD = 0;            /* CAN Controller is in the Operating Mode */
   
   pCANx->IER  = CxIER_RIE;   /* Enable Receiver Interrupt */
   pCANx->IER |= CxIER_EIE;   /* Enable Error warning Interrupt */
   pCANx->IER |= CxICR_DOI;   /* Enable Data Overrun Interrupt */
   
#if (_TALCPU_CAN_USE_AF >= 1)
   /* Enable Acceptance Filter */   
   AFMR = 0;
#else
   /* Set "Acceptance Filter Bypass Mode" */
   AFMR = AFMR_AccBP;
#endif
   
   Error = TAL_OK;
   
CANOpenEnd:   
   return(Error);
} /* cpu_CANOpen */

/*************************************************************************/
/*  cpu_CANClose                                                         */
/*                                                                       */
/*  Close the CAN port.                                                  */
/*                                                                       */
/*  In    : pDCB                                                         */
/*  Out   : none                                                         */
/*  Return: TAL_OK / error cause                                         */
/*************************************************************************/
TAL_RESULT cpu_CANClose (TAL_CAN_DCB *pDCB)
{
   TAL_RESULT    Error = TAL_OK;
   CAN_TypeDef *pCANx;
   
   /* Get CAN pointer */
   pCANx = (CAN_TypeDef*)pDCB->HW.dBaseAddress;

   pCANx->MOD = CxMOD_RM; /* CAN operation is disabled */
   pCANx->IER = 0;        /* Disable all interrupts */
   
   /* Clear filter and update acceptance filter */   
   memset(pDCB->HW.FilterArray, 0x00, sizeof(pDCB->HW.FilterArray));
   pDCB->HW.bFilterAllStd = TAL_FALSE;
   pDCB->HW.bFilterAllExt = TAL_FALSE;
   
   UpdateAcceptanceFilter();
      
   return(Error);
} /* cpu_CANClose */

/*************************************************************************/
/*  cpu_CANStartTx                                                       */
/*                                                                       */
/*  Send the data from the ring buffer if the TX interrupt is disabled.  */
/*                                                                       */
/*  In    : pDCB                                                         */
/*  Out   : none                                                         */
/*  Return: TAL_OK / error cause                                         */
/*************************************************************************/
TAL_RESULT cpu_CANStartTx (TAL_CAN_DCB *pDCB)
{
   TAL_RESULT       Error = TAL_ERROR;
   CAN_TypeDef    *pCANx;
   TAL_CAN_OBJECT   CANObject;
   
   /* Get CAN pointer */
   pCANx = (CAN_TypeDef *)pDCB->HW.dBaseAddress;

   TAL_CPU_DISABLE_ALL_INTS();
   if (pCANx->IER & CxIER_TIE1)
   {
      /* TX interrupt is enabled, do nothing */
   }
   else
   {
      Error = tal_MISCRingGet(&pDCB->TxRing, (uint8_t*)&CANObject);
      if (TAL_OK == Error)
      {
         Error = SendObject(pDCB, pCANx, &CANObject);
         if (TAL_OK == Error)
         {
            /* Enable TX interrupt */
            pCANx->IER |= CxIER_TIE1;
         }
      }
   }
   TAL_CPU_ENABLE_ALL_INTS();
   
   return(Error);
} /* cpu_CANStartTx */

/*************************************************************************/
/*  cpu_CANTxIsRunning                                                   */
/*                                                                       */
/*  Check if TX is still running.                                        */
/*                                                                       */
/*  In    : pDCB                                                         */
/*  Out   : none                                                         */
/*  Return: TAL_OK / TAL_ERROR                                           */
/*************************************************************************/
TAL_RESULT cpu_CANTxIsRunning (TAL_CAN_DCB *pDCB)
{
   TAL_RESULT  Error = TAL_OK;
   CAN_TypeDef *pCANx;
   
   /* Get CAN pointer */
   pCANx = (CAN_TypeDef *)pDCB->HW.dBaseAddress;

   TAL_CPU_DISABLE_ALL_INTS();
   if (pCANx->IER & CxIER_TIE1)
   {
      /* TX is still running */
      Error = TAL_OK;
   }
   else
   {
      /* TX is not running */
      Error = TAL_ERROR;
   }
   TAL_CPU_ENABLE_ALL_INTS();
   
   return(Error);
} /* cpu_CANTxIsRunning */

/*************************************************************************/
/*  cpu_CANEnableSilentMode                                              */
/*                                                                       */
/*  Enable "Silent" mode.                                                */
/*                                                                       */
/*  In    : pDCB                                                         */
/*  Out   : none                                                         */
/*  Return: TAL_OK / error cause                                         */
/*************************************************************************/
TAL_RESULT cpu_CANEnableSilentMode (TAL_CAN_DCB *pDCB)
{
   (void)pDCB;   
   
   /* Not supported */
   
   return(TAL_ERROR);
} /* cpu_CANEnableSilentMode */

/*************************************************************************/
/*  cpu_CANDisableSilentMode                                             */
/*                                                                       */
/*  Disable "Silent" mode.                                               */
/*                                                                       */
/*  In    : pDCB                                                         */
/*  Out   : none                                                         */
/*  Return: TAL_OK / error cause                                         */
/*************************************************************************/
TAL_RESULT cpu_CANDisableSilentMode (TAL_CAN_DCB *pDCB)
{
   (void)pDCB;   
   
   /* Not supported */
   
   return(TAL_ERROR);
} /* cpu_CANDisableSilentMode */

/*************************************************************************/
/*  cpu_CANIdentRegister                                                 */
/*                                                                       */
/*  Register a CAN identifier.                                           */
/*                                                                       */
/*  In    : pDCB, dIdentifier, Type                                      */
/*  Out   : none                                                         */
/*  Return: TAL_OK / error cause                                         */
/*************************************************************************/
TAL_RESULT cpu_CANIdentRegister (TAL_CAN_DCB  *pDCB,
                                uint32_t       dIdentifier,
                                TAL_CAN_TYPE_ID Type)
{
   TAL_RESULT  Error;

   Error = CANIdentRegister(pDCB, dIdentifier, Type);
   
   return(Error);
} /* cpu_CANIdentRegister */   

/*************************************************************************/
/*  cpu_CANIdentDeRegister                                               */
/*                                                                       */
/*  DeRegister a CAN identifier.                                         */
/*                                                                       */
/*  In    : pDCB, dIdentifier, Type                                      */
/*  Out   : none                                                         */
/*  Return: TAL_OK / error cause                                         */
/*************************************************************************/
TAL_RESULT cpu_CANIdentDeRegister (TAL_CAN_DCB  *pDCB,
                                  uint32_t       dIdentifier,
                                  TAL_CAN_TYPE_ID Type)
{
   TAL_RESULT  Error;

   Error = CANIdentDeRegister(pDCB, dIdentifier, Type);

   return(Error);
} /* cpu_CANIdentDeRegister */   

/*** EOF ***/
